import sys
import textwrap
try:
    import unittest2 as unittest
except ImportError:
    import unittest
from paramunittest import *
from radon.visitors import *


dedent = lambda code: textwrap.dedent(code).strip()


SIMPLE_BLOCKS = [
    ('''
     if a: pass
     ''', 2),

    ('''
     if a: pass
     else: pass
     ''', 2),

    ('''
     if a: pass
     elif b: pass
     ''', 3),

    ('''
     if a: pass
     elif b: pass
     else: pass
     ''', 3),

    ('''
    if a and b: pass
    ''', 3),

    ('''
    if a and b: pass
    else: pass
    ''', 3),

    ('''
     if a and b: pass
     elif c and d: pass
     else: pass
     ''', 5),

    ('''
     if a and b or c and d: pass
     else: pass
     ''', 5),

    ('''
     if a and b or c: pass
     else: pass
     ''', 4),

    ('''
     for x in range(10): print(x)
     ''', 2),

    ('''
     for x in xrange(10): print(x)
     else: pass
     ''', 3),

    ('''
     while a < 4: pass
     ''', 2),

    ('''
     while a < 4: pass
     else: pass
     ''', 3),

    ('''
     while a < 4 and b < 42: pass
     ''', 3),

    ('''
     while a and b or c < 10: pass
     else: pass
     ''', 5),

    ('''
     with open('raw.py') as fobj: print(fobj.read())
     ''', 2),

    ('''
     [i for i in range(4)]
     ''', 2),

    ('''
     [i for i in range(4) if i&1]
     ''', 3),

    ('''
     (i for i in range(4))
     ''', 2),

    ('''
     (i for i in range(4) if i&1)
     ''', 3),

    ('''
     [i for i in range(42) if sum(k ** 2 for k in divisors(i)) & 1]
     ''', 4),

    ('''
     try: raise TypeError
     except TypeError: pass
     ''', 2),

    ('''
     try: raise TypeError
     except TypeError: pass
     else: pass
     ''', 3),

    ('''
     try: raise TypeError
     finally: pass
     ''', 1),

    ('''
     try: raise TypeError
     except TypeError: pass
     finally: pass
     ''', 2),

    ('''
     try: raise TypeError
     except TypeError: pass
     else: pass
     finally: pass
     ''', 3),

    ('''
     k = lambda a, b: k(b, a)
     ''', 2),

    ('''
     v = a if b else c
     ''', 2),

    ('''
     v = a if sum(i for i in xrange(c)) < 10 else c
     ''', 3),

    ('''
     sum(i for i in range(12) for z in range(i ** 2) if i * z & 1)
     ''', 4),

    ('''
     sum(i for i in range(10) if i >= 2 and val and val2 or val3)
     ''', 6),

    ('''
     for i in range(10):
         print(i)
     else:
         print('wah')
         print('really not found')
         print(3)
     ''', 3),

    ('''
     while True:
         print(1)
     else:
         print(2)
         print(1)
         print(0)
         print(-1)
     ''', 3),

    ('''
     assert i < 0
     ''', 2),

    ('''
     assert i < 0, "Fail"
     ''', 2),

    ('''
     assert i < 0
     ''', 1, {'no_assert': True}),

    ('''
     def f():
        assert 10 > 20
     ''', 1, {'no_assert': True}),

    ('''
     class TestYo(object):
        def test_yo(self):
            assert self.n > 4
     ''', 1, {'no_assert': True}),
]


# These run only if Python version is >= 2.7
ADDITIONAL_BLOCKS = [
    ('''
     {i for i in range(4)}
     ''', 2),

    ('''
     {i for i in range(4) if i&1}
     ''', 3),

    ('''
     {i:i**4 for i in range(4)}
     ''', 2),

    ('''
     {i:i**4 for i in range(4) if i&1}
     ''', 3),
]

BLOCKS = SIMPLE_BLOCKS[:]
if sys.version_info[:2] >= (2, 7):
    BLOCKS.extend(ADDITIONAL_BLOCKS)

@parametrized(*BLOCKS)
class TestBlocks(ParametrizedTestCase):
    '''Test blocks.'''

    def setParameters(self, code, expected_complexity, kwargs={}):
        self.code = dedent(code)
        self.expected_complexity = expected_complexity
        self.kwargs = kwargs

    def testComplexityVisitor(self):
        visitor = ComplexityVisitor.from_code(self.code, **self.kwargs)
        self.assertEqual(visitor.complexity, self.expected_complexity)



SINGLE_FUNCTIONS_CASES = [
    ('''
     def f(a, b, c):
        if a and b == 4:
            return c ** c
        elif a and not c:
            return sum(i for i in range(41) if i&1)
        return a + b
     ''', (1, 7)),

    ('''
     if a and not b: pass
     elif b or c: pass
     else: pass

     for i in range(4):
        print(i)

     def g(a, b):
        while a < b:
            b, a = a **2, b ** 2
        return b
     ''', (6, 2)),

    ('''
     def f(a, b):
        while a**b:
            a, b = b, a * (b - 1)
            if a and b:
                b = 0
            else:
                b = 1
        return sum(i for i in range(b))
     ''', (1, 5)),
]


@parametrized(*SINGLE_FUNCTIONS_CASES)
class TestSingleFunctions(ParametrizedTestCase):

    def setParameters(self, code, expected_complexity):
        self.code = dedent(code)
        self.expected_complexity = expected_complexity

    def testComplexityVisitor(self):
        visitor = ComplexityVisitor.from_code(self.code)
        self.assertEqual(len(visitor.functions), 1)
        self.assertEqual((visitor.complexity, visitor.functions[0].complexity),
                         self.expected_complexity)


FUNCTIONS_CASES = [
    ('''
     def f(a, b):
        return a if b else 2

     def g(a, b, c):
        if a and b:
            return a / b + b / a
        elif b and c:
            return b / c - c / b
        return a + b + c

     def h(a, b):
        return 2 * (a + b)
     ''', (2, 5, 1)),

    ('''
     def f(p, q):
        while p:
            p, q = q, p - q
        if q < 1:
            return 1 / q ** 2
        elif q > 100:
            return 1 / q ** .5
        return 42 if not q else p

     def g(a, b, c):
        if a and b or a - b:
            return a / b - c
        elif b or c:
            return 1
        else:
            k = 0
            with open('results.txt', 'w') as fobj:
                for i in range(b ** c):
                    k += sum(1 / j for j in range(i ** 2) if j > 2)
                fobj.write(str(k))
            return k - 1
     ''', (5, 10)),
]


@parametrized(*FUNCTIONS_CASES)
class TestFunctions(ParametrizedTestCase):

    def setParameters(self, code, expected_complexity):
        self.code = dedent(code)
        self.expected_complexity = expected_complexity

    def testComplexityVisitor(self):
        visitor = ComplexityVisitor.from_code(self.code)
        self.assertEqual(len(visitor.functions), len(self.expected_complexity))
        self.assertEqual(tuple(map(GET_COMPLEXITY, visitor.functions)),
                         self.expected_complexity)


CLASSES_CASES = [
    ('''
     class A(object):

         def m(self, a, b):
             if not a or b:
                 return b - 1
             try:
                 return a / b
             except ZeroDivisionError:
                 return a

         def n(self, k):
             while self.m(k) < k:
                 k -= self.m(k ** 2 - min(self.m(j) for j in range(k ** 4)))
             return k
     ''', (6, 4, 3)),

    ('''
     class B(object):

         ATTR = 9 if A().n(9) == 9 else 10
         import sys
         if sys.version_info >= (3, 3):
             import os
             AT = os.openat('/random/loc')

         def __iter__(self):
             return __import__('itertools').tee(B.__dict__)

         def test(self, func):
             a = func(self.ATTR, self.AT)
             if a < self.ATTR:
                 yield self
             elif a > self.ATTR ** 2:
                 yield self.__iter__()
             yield iter(a)
     ''', (5, 1, 3)),
]


@parametrized(*CLASSES_CASES)
class TestClasses(ParametrizedTestCase):

    def setParameters(self, code, expected_complexity):
        self.code = dedent(code)
        self.total_class_complexity = expected_complexity[0]
        self.methods_complexity = expected_complexity[1:]

    def testComplexityVisitor(self):
        visitor = ComplexityVisitor.from_code(self.code)
        self.assertEqual(len(visitor.classes), 1)
        self.assertEqual(len(visitor.functions), 0)
        cls = visitor.classes[0]
        self.assertEqual(cls.real_complexity, self.total_class_complexity)
        self.assertEqual(tuple(map(GET_COMPLEXITY, cls.methods)),
                         self.methods_complexity)


GENERAL_CASES = [
    ('''
     if a and b:
         print
     else:
         print
     a = sum(i for i in range(1000) if i % 3 == 0 and i % 5 == 0)

     def f(n):
         if n == 0:
             return 1
         elif n == 1:
             return n
         elif n < 5:
             return (n - 1) ** 2
         return n * pow(n, f(n - 1), n - 3)
     ''', (6, 3, 0, 9)),

    ('''
     try:
         1 / 0
     except ZeroDivisonError:
         print
     except TypeError:
         pass

     class J(object):

         def aux(self, w):
             if w == 0:
                 return 0
             return w - 1 + sum(self.aux(w - 3 - i) for i in range(2))

     def f(a, b):
         if a < b:
             b, a = a, b
         return a, b
     ''', (3, 1, 2, 6)),
]


@parametrized(*GENERAL_CASES)
class TestModules(ParametrizedTestCase):

    def setParameters(self, code, expected_complexity):
        self.code = dedent(code)
        self.module_complexity, self.functions_complexity, \
            self.classes_complexity, \
            self.total_complexity = expected_complexity

    def testModule(self):
        visitor = ComplexityVisitor.from_code(self.code)
        self.assertEqual(visitor.complexity, self.module_complexity)
        self.assertEqual(visitor.functions_complexity,
                         self.functions_complexity)
        self.assertEqual(visitor.classes_complexity, self.classes_complexity)
        self.assertEqual(visitor.total_complexity, self.total_complexity)


CLOJURES_CASES = [
    ('''
     def f(n):
         def g(l):
             return l ** 4
         def h(i):
             return i ** 5 + 1 if i & 1 else 2
         return sum(g(u + 4) / float(h(u)) for u in range(2, n))
     ''', ('g', 'h'), (1, 2, 3)),

    ('''
     # will it work? :D
     def memoize(func):
         cache = {}
         def aux(*args, **kwargs):
             key = (args, kwargs)
             if key in cache:
                 return cache[key]
             cache[key] = res = func(*args, **kwargs)
             return res
         return aux
     ''', ('aux',), (2, 2)),
]


@parametrized(*CLOJURES_CASES)
class TestClojures(ParametrizedTestCase):

    def setParameters(self, code, clojure_names, expected_cc):
        self.visitor = ComplexityVisitor.from_code(dedent(code))
        self.func = self.visitor.functions[0]
        self.clojure_names = clojure_names
        self.expected_cj_cc = expected_cc[:-1]
        self.expected_total_cc = expected_cc[-1]

    def testOneFunction(self):
        self.assertEqual(len(self.visitor.functions), 1)

    def testClojureNames(self):
        names = tuple(cj.name for cj in self.func.clojures)
        self.assertEqual(names, self.clojure_names)

    def testClojuresComplexity(self):
        cj_complexity = tuple(cj.complexity for cj in self.func.clojures)
        self.assertEqual(cj_complexity, self.expected_cj_cc)

    def testTotalComplexity(self):
        self.assertEqual(self.func.complexity, self.expected_total_cc)

    def testMutableBlocks(self):
        # There was a bug for which `blocks` increased while it got accessed
        v = self.visitor
        self.assertTrue(v.blocks == v.blocks == v.blocks)


CONTAINERS_CASES = [
    (('func', 12, 0, 18, False, None, [], 5),
     ('F', 'func', 'F 12:0->18 func - 5')),

    (('meth', 12, 0, 21, True, 'cls', [], 5),
     ('M', 'cls.meth', 'M 12:0->21 cls.meth - 5')),

    (('cls', 12, 0, 15, [], 5),
     ('C', 'cls', 'C 12:0->15 cls - 5')),

    (('cls', 12, 0, 19, [object, object, object, object], 30),
     ('C', 'cls', 'C 12:0->19 cls - 8')),
]


@parametrized(*CONTAINERS_CASES)
class TestContainers(ParametrizedTestCase):

    def setParameters(self, values, expected):
        self.values = values
        self.expected_letter = expected[0]
        self.expected_name = expected[1]
        self.expected_str = expected[2]

    def testContainers(self):
        cls = Function if len(self.values) == 8 else Class
        obj = cls(*self.values)
        self.assertEqual(obj.letter, self.expected_letter)
        self.assertEqual(obj.fullname, self.expected_name)
        self.assertEqual(str(obj), self.expected_str)



import unittest
import collections
from radon.cli import Config
import radon.cli.harvest as harvest
try:
    import unittest.mock as mock
except ImportError:
    import mock

BASE_CONFIG = Config(
    exclude='test_[^.]+\.py',
    ignore='tests,docs',
)

CC_CONFIG = Config(
    order='SCORE',
    no_assert=False,
    min='A',
    max='F',
    show_complexity=False,
    average=True,
    total_average=False,
    **BASE_CONFIG.config_values
)

RAW_CONFIG = Config(
    summary=True,
)

MI_CONFIG = Config(
    multi=True,
    min='B',
    max='C',
    show=True,
)


def fake_gobble(fobj):
    return 42


def fake_gobble_raising(fobj):
    raise TypeError('mystr')


def fake_run():
    for i in range(3):
        yield {'file-{0}'.format(i): i**2}


class ConfigMixin(object):

    def setUp(self):
        self.config = Config(**self.CONFIG_CLASS.config_values.copy())


class TestBaseHarvester(ConfigMixin, unittest.TestCase):

    CONFIG_CLASS = BASE_CONFIG

    @mock.patch('radon.cli.harvest.iter_filenames')
    def test_iter_filenames(self, iter_mock):
        h = harvest.Harvester([], self.config)
        h._iter_filenames()

        iter_mock.assert_called_with([], self.config.exclude,
                                     self.config.ignore)

    def test_gobble_not_implemented(self):
        h = harvest.Harvester([], self.config)
        self.assertRaises(NotImplementedError, h.gobble, None)

    def test_as_xml_not_implemented(self):
        h = harvest.Harvester([], self.config)
        self.assertRaises(NotImplementedError, h.as_xml)

    def test_to_terminal_not_implemented(self):
        h = harvest.Harvester([], self.config)
        self.assertRaises(NotImplementedError, h.to_terminal)

    def test_run(self):
        h = harvest.Harvester(['-'], self.config)
        h.gobble = fake_gobble
        self.assertTrue(isinstance(h.run(), collections.Iterator))
        self.assertEqual(list(h.run()), [('-', 42)])
        h.gobble = fake_gobble_raising
        self.assertEqual(list(h.run()), [('-', {'error': 'mystr'})])

    def test_results(self):
        h = harvest.Harvester([], self.config)
        h.run = fake_run
        results = h.results
        self.assertTrue(isinstance(results, collections.Iterator))
        self.assertEqual(list(results), [{'file-0': 0}, {'file-1': 1},
                                         {'file-2': 4}])
        self.assertFalse(isinstance(h.results, collections.Iterator))
        self.assertTrue(isinstance(h.results, collections.Iterable))
        self.assertTrue(isinstance(h.results, list))

    def test_as_json(self):
        h = harvest.Harvester([], self.config)
        h._results = {'filename': {'complexity': 2}}
        self.assertEqual(h.as_json(), '{"filename": {"complexity": 2}}')


class TestCCHarvester(ConfigMixin, unittest.TestCase):

    CONFIG_CLASS = CC_CONFIG

    @mock.patch('radon.cli.harvest.sorted_results')
    @mock.patch('radon.cli.harvest.cc_visit')
    def test_gobble(self, cc_mock, sr_mock):
        cc_mock.return_value = mock.sentinel.one
        fobj = mock.MagicMock()
        fobj.read.return_value = mock.sentinel.two

        h = harvest.CCHarvester([], self.config)
        h.gobble(fobj)

        self.assertTrue(fobj.read.called)
        cc_mock.assert_called_with(mock.sentinel.two,
                                   no_assert=self.config.no_assert)
        sr_mock.assert_called_with(mock.sentinel.one, order=self.config.order)

    @mock.patch('radon.cli.harvest.cc_to_dict')
    def test_to_dicts(self, c2d_mock):
        c2d_mock.side_effect = lambda i: i
        h = harvest.CCHarvester([], self.config)
        sample_results = [('a', [{'rank': 'A'}]), ('b', [{'rank': 'B'}]),
                          ('c', {'error': 'An ERROR!'})]
        h._results = sample_results

        self.assertEqual(h._to_dicts(), dict(sample_results))
        self.assertEqual(c2d_mock.call_count, 2)

        h.config.min = 'B'
        h._results = sample_results[1:]
        self.assertEqual(h._to_dicts(), dict(sample_results[1:]))

    @mock.patch('radon.cli.harvest.dict_to_xml')
    def test_as_json_xml(self, d2x_mock):
        to_dicts_mock = mock.MagicMock()
        to_dicts_mock.return_value = {'a': {'rank': 'A'}}

        h = harvest.CCHarvester([], self.config)
        h._to_dicts = to_dicts_mock
        self.assertEqual(h.as_json(), '{"a": {"rank": "A"}}')

        h.as_xml()
        self.assertTrue(d2x_mock.called)
        d2x_mock.assert_called_with(to_dicts_mock.return_value)
        self.assertEqual(to_dicts_mock.call_count, 2)

    @mock.patch('radon.cli.harvest.RESET')
    @mock.patch('radon.cli.harvest.RANKS_COLORS')
    @mock.patch('radon.cli.harvest.cc_to_terminal')
    def test_to_terminal(self, c2t_mock, ranks_mock, reset_mock):
        h = harvest.CCHarvester([], self.config)
        h._results = [
            ('a', {'error': 'mystr'}), ('b', {})
        ]
        c2t_mock.return_value = (['res'], 9, 3)
        ranks_mock.__getitem__.return_value = '<|A|>'
        reset_mock.__eq__.side_effect = lambda o: o == '__R__'

        results = list(h.to_terminal())
        c2t_mock.assert_called_once_with({}, self.config.show_complexity,
                                         self.config.min, self.config.max,
                                         self.config.total_average)
        self.assertEqual(results, [
            ('a', ('mystr',), {'error': True}),
            ('b', (), {}),
            (['res'], (), {'indent': 1}),
            ('\n{0} blocks (classes, functions, methods) analyzed.', (3,), {}),
            ('Average complexity: {0}{1} ({2}){3}',
             ('<|A|>', 'A', 3, '__R__'), {}),
        ])


class TestRawHarvester(ConfigMixin, unittest.TestCase):

    CONFIG_CLASS = RAW_CONFIG

    @mock.patch('radon.cli.harvest.raw_to_dict')
    @mock.patch('radon.cli.harvest.analyze')
    def test_gobble(self, analyze_mock, r2d_mock):
        fobj = mock.MagicMock()
        fobj.read.return_value = mock.sentinel.one
        analyze_mock.return_value = mock.sentinel.two

        h = harvest.RawHarvester([], self.config)
        h.gobble(fobj)

        self.assertEqual(fobj.read.call_count, 1)
        analyze_mock.assert_called_once_with(mock.sentinel.one)
        r2d_mock.assert_called_once_with(mock.sentinel.two)

    def test_as_xml(self):
        h = harvest.RawHarvester([], self.config)
        self.assertRaises(NotImplementedError, h.as_xml)

    def test_to_terminal(self):
        h = harvest.RawHarvester([], self.config)
        h._results = [
            ('a', {'error': 'mystr'}),
            ('b', {'loc': 24, 'lloc': 27, 'sloc': 15, 'comments': 3,
                   'multi': 3, 'blank': 9}),
            ('c', {'loc': 24, 'lloc': 27, 'sloc': 15, 'comments': 3,
                   'multi': 3, 'blank': 9}),
            ('e', {'loc': 0, 'lloc': 0, 'sloc': 0, 'comments': 0, 'multi': 0,
                   'blank': 0}),
        ]

        self.assertEqual(list(h.to_terminal()), [
            ('a', ('mystr',), {'error': True}),
            ('b', (), {}),
            ('{0}: {1}', ('LOC', 24), {'indent': 1}),
            ('{0}: {1}', ('LLOC', 27), {'indent': 1}),
            ('{0}: {1}', ('SLOC', 15), {'indent': 1}),
            ('{0}: {1}', ('Comments', 3), {'indent': 1}),
            ('{0}: {1}', ('Multi', 3), {'indent': 1}),
            ('{0}: {1}', ('Blank', 9), {'indent': 1}),
            ('- Comment Stats', (), {'indent': 1}),
            ('(C % L): {0:.0%}', (0.125,), {'indent': 2}),
            ('(C % S): {0:.0%}', (0.2,), {'indent': 2}),
            ('(C + M % L): {0:.0%}', (0.25,), {'indent': 2}),
            ('c', (), {}),
            ('{0}: {1}', ('LOC', 24), {'indent': 1}),
            ('{0}: {1}', ('LLOC', 27), {'indent': 1}),
            ('{0}: {1}', ('SLOC', 15), {'indent': 1}),
            ('{0}: {1}', ('Comments', 3), {'indent': 1}),
            ('{0}: {1}', ('Multi', 3), {'indent': 1}),
            ('{0}: {1}', ('Blank', 9), {'indent': 1}),
            ('- Comment Stats', (), {'indent': 1}),
            ('(C % L): {0:.0%}', (0.125,), {'indent': 2}),
            ('(C % S): {0:.0%}', (0.2,), {'indent': 2}),
            ('(C + M % L): {0:.0%}', (0.25,), {'indent': 2}),
            ('e', (), {}),
            ('{0}: {1}', ('LOC', 0), {'indent': 1}),
            ('{0}: {1}', ('LLOC', 0), {'indent': 1}),
            ('{0}: {1}', ('SLOC', 0), {'indent': 1}),
            ('{0}: {1}', ('Comments', 0), {'indent': 1}),
            ('{0}: {1}', ('Multi', 0), {'indent': 1}),
            ('{0}: {1}', ('Blank', 0), {'indent': 1}),
            ('- Comment Stats', (), {'indent': 1}),
            ('(C % L): {0:.0%}', (0,), {'indent': 2}),
            ('(C % S): {0:.0%}', (0,), {'indent': 2}),
            ('(C + M % L): {0:.0%}', (0,), {'indent': 2}),
            ('** Total **', (), {}),
            ('{0}: {1}', ('LOC', 48), {'indent': 1}),
            ('{0}: {1}', ('LLOC', 54), {'indent': 1}),
            ('{0}: {1}', ('SLOC', 30), {'indent': 1}),
            ('{0}: {1}', ('Comments', 6), {'indent': 1}),
            ('{0}: {1}', ('Multi', 6), {'indent': 1}),
            ('{0}: {1}', ('Blank', 18), {'indent': 1}),
        ])


class TestMIHarvester(ConfigMixin, unittest.TestCase):

    CONFIG_CLASS = MI_CONFIG

    @mock.patch('radon.cli.harvest.mi_visit')
    def test_gobble(self, mv_mock):
        fobj = mock.MagicMock()
        fobj.read.return_value = mock.sentinel.one
        mv_mock.return_value = 23.5

        h = harvest.MIHarvester([], self.config)
        result = h.gobble(fobj)

        self.assertEqual(fobj.read.call_count, 1)
        mv_mock.assert_called_once_with(mock.sentinel.one, self.config.multi)
        self.assertEqual(result, {'mi': 23.5, 'rank': 'A'})

    @mock.patch('radon.cli.harvest.json.dumps')
    def test_as_json(self, d_mock):
        h = harvest.MIHarvester([], self.config)
        h.config.min = 'C'
        h._results = [
            ('a', {'error': 'mystr'}),
            ('b', {'mi': 25, 'rank': 'A'}),
            ('c', {'mi': 15, 'rank': 'B'}),
            ('d', {'mi': 0, 'rank': 'C'}),
        ]

        h.as_json()
        d_mock.assert_called_with(dict([h._results[0], h._results[-1]]))

    def test_as_xml(self):
        h = harvest.MIHarvester([], self.config)
        self.assertRaises(NotImplementedError, h.as_xml)

    @mock.patch('radon.cli.harvest.RESET')
    @mock.patch('radon.cli.harvest.MI_RANKS')
    def test_to_terminal(self, ranks_mock, reset_mock):
        ranks_mock.__getitem__.side_effect = lambda j: '<|{0}|>'.format(j)
        reset_mock.__eq__.side_effect = lambda o: o == '__R__'

        h = harvest.MIHarvester([], self.config)
        h._results = [
            ('a', {'error': 'mystr'}),
            ('b', {'mi': 25, 'rank': 'A'}),
            ('c', {'mi': 15, 'rank': 'B'}),
            ('d', {'mi': 0, 'rank': 'C'}),
        ]

        self.assertEqual(list(h.to_terminal()), [
            ('a', ('mystr',), {'error': True}),
            ('{0} - {1}{2}{3}{4}', ('c', '<|B|>', 'B', ' (15.00)', '__R__'),
             {}),
            ('{0} - {1}{2}{3}{4}', ('d', '<|C|>', 'C', ' (0.00)', '__R__'),
             {}),
        ])



'''This module contains various utility functions used in the CLI interface.'''

import os
import sys
import fnmatch
import xml.etree.cElementTree as et
from contextlib import contextmanager
from radon.visitors import Function
from radon.complexity import cc_rank
from radon.cli.colors import (LETTERS_COLORS, RANKS_COLORS, TEMPLATE, BRIGHT,
                              RESET)


@contextmanager
def _open(path):
    '''Mock of the built-in `open()` function. If `path` is `-` then
    `sys.stdin` is returned.
    '''
    if path == '-':
        yield sys.stdin
    else:
        with open(path) as f:
            yield f


def iter_filenames(paths, exclude=None, ignore=None):
    '''A generator that yields all sub-paths of the ones specified in
    `paths`. Optional `exclude` filters can be passed as a comma-separated
    string of regexes, while `ignore` filters are a comma-separated list of
    directory names to ignore. Ignore patterns are can be plain names or glob
    patterns. If paths contains only a single hyphen, stdin is implied,
    returned as is.
    '''
    if set(paths) == set(('-',)):
        yield '-'
        return
    for path in paths:
        if os.path.isfile(path):
            yield path
            continue
        for filename in explore_directories(path, exclude, ignore):
            yield filename


def explore_directories(start, exclude, ignore):
    '''Explore files and directories under `start`. `explore` and `ignore`
    arguments are the same as in :func:`iter_filenames`.
    '''
    e = '*[!p][!y]'
    exclude = '{0},{1}'.format(e, exclude).split(',') if exclude else [e]
    ignore = '.*,{0}'.format(ignore).split(',') if ignore else ['.*']
    for root, dirs, files in os.walk(start):
        dirs[:] = list(filter_out(dirs, ignore))
        fullpaths = (os.path.normpath(os.path.join(root, p)) for p in files)
        for filename in filter_out(fullpaths, exclude):
            if (not os.path.basename(filename).startswith('.') and
                    filename.endswith('.py')):
                yield filename


def filter_out(strings, patterns):
    '''Filter out any string that matches any of the specified patterns.'''
    for s in strings:
        if all(not fnmatch.fnmatch(s, p) for p in patterns):
            yield s


def cc_to_dict(obj):
    '''Convert an object holding CC results into a dictionary. This is meant
    for JSON dumping.'''
    def get_type(obj):
        '''The object can be of type *method*, *function* or *class*.'''
        if isinstance(obj, Function):
            return 'method' if obj.is_method else 'function'
        return 'class'

    result = {
        'type': get_type(obj),
        'rank': cc_rank(obj.complexity),
    }
    attrs = set(Function._fields) - set(('is_method', 'clojures'))
    for a in attrs:
        v = getattr(obj, a, None)
        if v is not None:
            result[a] = v
    for key in ('methods', 'clojures'):
        if hasattr(obj, key):
            result[key] = list(map(cc_to_dict, getattr(obj, key)))
    return result


def raw_to_dict(obj):
    '''Convert an object holding raw analysis results into a dictionary. This
    is meant for JSON dumping.'''
    result = {}
    for a in obj._fields:
        v = getattr(obj, a, None)
        if v is not None:
            result[a] = v
    return result


def dict_to_xml(results):
    '''Convert a dictionary holding CC analysis result into a string containing
    xml.'''
    ccm = et.Element('ccm')
    for filename, blocks in results.items():
        for block in blocks:
            metric = et.SubElement(ccm, 'metric')
            complexity = et.SubElement(metric, 'complexity')
            complexity.text = str(block['complexity'])
            unit = et.SubElement(metric, 'unit')
            name = block['name']
            if 'classname' in block:
                name = '{0}.{1}'.format(block['classname'], block['name'])
            unit.text = name
            classification = et.SubElement(metric, 'classification')
            classification.text = block['rank']
            file = et.SubElement(metric, 'file')
            file.text = filename
    return et.tostring(ccm).decode('utf-8')


def cc_to_terminal(results, show_complexity, min, max, total_average):
    '''Transfom Cyclomatic Complexity results into a 3-elements tuple:

        ``(res, total_cc, counted)``

    `res` is a list holding strings that are specifically formatted to be
    printed to a terminal.
    `total_cc` is a number representing the total analyzed cyclomatic
    complexity.
    `counted` holds the number of the analyzed blocks.

    If *show_complexity* is `True`, then the complexity of a block will be
    shown in the terminal line alongside its rank.
    *min* and *max* are used to control which blocks are shown in the resulting
    list. A block is formatted only if its rank is `min <= rank <= max`.
    If *total_average* is `True`, the `total_cc` and `counted` count every
    block, regardless of the fact that they are formatted in `res` or not.
    '''
    res = []
    counted = 0
    total_cc = .0
    for line in results:
        ranked = cc_rank(line.complexity)
        if min <= ranked <= max:
            total_cc += line.complexity
            counted += 1
            res.append(_format_line(line, ranked, show_complexity))
        elif total_average:
            total_cc += line.complexity
            counted += 1
    return res, total_cc, counted


def _format_line(block, ranked, show_complexity=False):
    '''Format a single block as a line.

    *ranked* is the rank given by the `~radon.complexity.rank` function. If
    *show_complexity* is True, then the complexity score is added alongside.
    '''
    letter_colored = LETTERS_COLORS[block.letter] + block.letter
    rank_colored = RANKS_COLORS[ranked] + ranked
    compl = '' if not show_complexity else ' ({0})'.format(block.complexity)
    return TEMPLATE.format(BRIGHT, letter_colored, block.lineno,
                           block.col_offset, block.fullname, rank_colored,
                           compl, reset=RESET)



import unittest
import radon.cli as cli
import radon.complexity as cc_mod
from radon.cli.harvest import Harvester
try:
    import unittest.mock as mock
except ImportError:
    import mock


def func(a, b=2, c=[], d=None):
    pass


def func2(*args, **kwargs):
    pass


def func3(b=3, *args):
    pass


def fake_to_terminal():
    yield ('a', ('mystr',), {'error': True})
    yield ('b', (), {})
    yield (('p1', 'p2'), (), {'indent': 1})


class TestConfig(unittest.TestCase):

    def test_base_behavior(self):
        c = cli.Config(a=2, b=3)
        self.assertEqual(c.config_values, {'a': 2, 'b': 3})
        self.assertEqual(c.a, 2)
        self.assertEqual(c.b, 3)

    def test_exceptions(self):
        c = cli.Config(a=2)
        self.assertEqual(c.__dict__, {'config_values': {'a': 2}})
        self.assertRaises(AttributeError, lambda: c.notexistent)

    def test_str(self):
        self.assertEqual(str(cli.Config()), '{}')
        self.assertEqual(str(cli.Config(a=2)), '{\'a\': 2}')

    def test_eq(self):
        self.assertEqual(cli.Config(), cli.Config())
        self.assertEqual(cli.Config(a=2), cli.Config(a=2))
        self.assertNotEqual(cli.Config(a=2), cli.Config(b=2))

    def test_for(self):
        self.assertEqual(cli.Config.from_function(func),
                         cli.Config(b=2, c=[], d=None))
        self.assertEqual(cli.Config.from_function(func2), cli.Config())
        self.assertEqual(cli.Config.from_function(func3), cli.Config(b=3))


class TestCommands(unittest.TestCase):

    @mock.patch('radon.cli.log_result')
    @mock.patch('radon.cli.CCHarvester')
    def test_cc(self, harv_mock, log_mock):
        harv_mock.return_value = mock.sentinel.harvester

        cli.cc(['-'], json=True)

        harv_mock.assert_called_once_with(['-'], cli.Config(
            min='A', max='F', exclude=None, ignore=None, show_complexity=False,
            average=False, order=getattr(cc_mod, 'SCORE'), no_assert=False,
            total_average=False))
        log_mock.assert_called_once_with(mock.sentinel.harvester, json=True,
                                         xml=False)

    @mock.patch('radon.cli.log_result')
    @mock.patch('radon.cli.RawHarvester')
    def test_raw(self, harv_mock, log_mock):
        harv_mock.return_value = mock.sentinel.harvester

        cli.raw(['-'], summary=True, json=True)

        harv_mock.assert_called_once_with(['-'], cli.Config(exclude=None,
                                                            ignore=None,
                                                            summary=True))
        log_mock.assert_called_once_with(mock.sentinel.harvester, json=True)

    @mock.patch('radon.cli.log_result')
    @mock.patch('radon.cli.MIHarvester')
    def test_mi(self, harv_mock, log_mock):
        harv_mock.return_value = mock.sentinel.harvester

        cli.mi(['-'], show=True, multi=False)

        harv_mock.assert_called_once_with(['-'], cli.Config(
            min='A', max='C', exclude=None, ignore=None, show=True,
            multi=False))
        log_mock.assert_called_once_with(mock.sentinel.harvester, json=False)


@mock.patch('radon.cli.sys.stdout.write')
class TestLogging(unittest.TestCase):

    def test_log(self, stdout_mock):
        cli.log('msg')
        cli.log('msg', indent=1)
        cli.log('{0} + 1', 2)
        cli.log('{0} + 1', 2, noformat=True)

        stdout_mock.assert_has_calls([
            mock.call('msg\n'),
            mock.call('    msg\n'),
            mock.call('2 + 1\n'),
            mock.call('{0} + 1\n'),
        ])
        self.assertEqual(stdout_mock.call_count, 4)

    def test_log_list(self, stdout_mock):
        cli.log_list([])
        cli.log_list(['msg'])

        stdout_mock.assert_called_once_with('msg\n')

    @mock.patch('radon.cli.RESET')
    @mock.patch('radon.cli.RED')
    @mock.patch('radon.cli.BRIGHT')
    def test_log_error(self, bright_mock, red_mock, reset_mock, stdout_mock):
        bright_mock.__str__.return_value = '@'
        red_mock.__str__.return_value = '<|||>'
        reset_mock.__str__.return_value = '__R__'

        cli.log_error('mystr')

        stdout_mock.assert_called_once_with('@<|||>ERROR__R__: mystr\n')

    @mock.patch('radon.cli.log_error')
    @mock.patch('radon.cli.log_list')
    @mock.patch('radon.cli.log')
    def test_log_result(self, log_mock, ll_mock, le_mock, stdout_mock):
        h = mock.Mock(spec=Harvester)
        h.as_json.return_value = mock.sentinel.json
        h.as_xml.return_value = mock.sentinel.xml
        h.to_terminal.side_effect = fake_to_terminal

        cli.log_result(h, json=True)
        h.as_json.assert_called_once_with()

        cli.log_result(h, json=True, xml=True)
        h.as_json.assert_called_twice_with()
        self.assertEqual(h.as_xml.call_count, 0)

        cli.log_result(h, xml=True)
        h.as_xml.assert_called_once_with()

        cli.log_result(h)
        h.to_terminal.assert_called_once_with()

        log_mock.assert_has_calls([
            mock.call(mock.sentinel.json, noformat=True),
            mock.call(mock.sentinel.json, noformat=True),
            mock.call(mock.sentinel.xml, noformat=True),
            mock.call('a'),
        ])
        le_mock.assert_called_once_with('mystr', indent=1)
        ll_mock.assert_has_calls([
            mock.call(['b']),
            mock.call(('p1', 'p2'), indent=1),
        ])



import os
import sys
import unittest
import radon.cli.tools as tools
from radon.visitors import Function, Class
from radon.raw import Module
try:
    import unittest.mock as mock
except ImportError:
    import mock
from paramunittest import parametrized


def fake_isfile(filename):
    if filename == 'file.py':
        return True
    return False


def fake_walk(start):
    dirs = ['tests', 'sub', '.hid']
    contents = {'tests': ['test_amod.py', 'run.py', '.hid.py'],
                'sub': ['amod.py', 'bmod.py']}
    yield '.', dirs, ['tox.ini', 'amod.py', 'test_all.py', 'fake.yp', 'noext']
    for d in dirs:
        yield './{0}'.format(d), [], contents[d]


class TestGenericTools(unittest.TestCase):

    def test_open(self):
        with tools._open('-') as fobj:
            self.assertTrue(fobj is sys.stdin)

        m = mock.mock_open()
        with mock.patch('radon.cli.tools.open', m, create=True):
            tools._open('randomfile.py').__enter__()
        m.assert_called_with('randomfile.py')


class TestIterFilenames(unittest.TestCase):

    def setUp(self):
        self.iter_files = lambda *a, **kw: list(tools.iter_filenames(*a, **kw))

    def assertPEqual(self, a, b):
        paths = [list(map(os.path.normpath, p)) for p in (a, b)]
        self.assertEqual(*paths)

    def test_stdin(self):
        self.assertEqual(self.iter_files(['-']), ['-'])

    @mock.patch('radon.cli.tools.os.path')
    @mock.patch('radon.cli.tools.os')
    def test_all(self, os_mod, os_path_mod):
        os_path_mod.normpath = os.path.normpath
        os_path_mod.basename = os.path.basename
        os_path_mod.join = os.path.join
        os_path_mod.isfile.side_effect = fake_isfile
        os_mod.walk = fake_walk

        self.assertPEqual(self.iter_files(['file.py', 'random/path']),
                          ['file.py', 'amod.py', 'test_all.py',
                           'tests/test_amod.py', 'tests/run.py', 'sub/amod.py',
                           'sub/bmod.py'])

        self.assertPEqual(self.iter_files(['file.py', 'random/path'],
                                          'test_*'),
                          ['file.py', 'amod.py', 'tests/test_amod.py',
                           'tests/run.py', 'sub/amod.py', 'sub/bmod.py'])

        self.assertPEqual(self.iter_files(['file.py', 'random/path'],
                                          '*test_*'),
                          ['file.py', 'amod.py', 'tests/run.py', 'sub/amod.py',
                           'sub/bmod.py'])

        self.assertPEqual(self.iter_files(['file.py', 'random/path'],
                                          '*/test_*,amod*'),
                          ['file.py', 'test_all.py', 'tests/run.py',
                           'sub/amod.py', 'sub/bmod.py'])

        self.assertPEqual(self.iter_files(['file.py', 'random/path'], None,
                                          'tests'),
                          ['file.py', 'amod.py', 'test_all.py', 'sub/amod.py',
                           'sub/bmod.py'])

        self.assertPEqual(self.iter_files(['file.py', 'random/path'], None,
                                          'tests,sub'),
                          ['file.py', 'amod.py', 'test_all.py'])


CC_RESULTS_CASES = [
    ([
        Function('name', 12, 0, 16, False, None, [], 6),
    ], {
        'type': 'function', 'name': 'name', 'lineno': 12, 'col_offset': 0,
        'endline': 16, 'clojures': [], 'complexity': 6, 'rank': 'B',
    }),
    ([
        Class('Classname', 17, 0, 29, [
            Function('name', 19, 4, 26, True, 'Classname', [], 7),
        ], 7),
    ], {
        'type': 'class', 'name': 'Classname', 'lineno': 17, 'col_offset': 0,
        'endline': 29, 'complexity': 8, 'rank': 'B', 'methods': [
            {
                'type': 'method', 'lineno': 19, 'col_offset': 4, 'endline': 26,
                'clojures': [], 'complexity': 7, 'rank': 'B', 'classname':
                'Classname', 'name': 'name',
            }
        ],
    }),
    ([
        Function('name', 12, 0, 16, False, None, [
            Function('aux', 13, 4, 17, False, None, [], 4),
        ], 10),
    ], {
        'type': 'function', 'name': 'name', 'lineno': 12, 'col_offset': 0,
        'endline': 16, 'complexity': 10, 'rank': 'B', 'clojures': [
            {
                'name': 'aux', 'lineno': 13, 'col_offset': 4, 'endline': 17,
                'clojures': [], 'complexity': 4, 'rank': 'A', 'type':
                'function',
            }
        ]
    }),
]


@parametrized(*CC_RESULTS_CASES)
class TestCCToDict(unittest.TestCase):

    def setParameters(self, blocks, **dict_result):
        self.blocks = blocks
        self.dict_result = dict_result

    def testCCToDict(self):
        self.assertEqual(tools.cc_to_dict(self.blocks), self.dict_result)


CC_TO_XML_CASE = [
    {'clojures': [], 'endline': 16, 'complexity': 6, 'lineno': 12, 'is_method':
     False, 'name': 'name', 'col_offset': 0, 'rank': 'B'},

    {'complexity': 8, 'endline': 29, 'rank': 'B', 'lineno': 17, 'name':
     'Classname', 'col_offset': 0},

    {'classname': 'Classname', 'clojures': [], 'endline': 26, 'complexity': 7,
     'lineno': 19, 'is_method': True, 'name': 'name', 'col_offset': 4,
     'rank': 'B'},

    {'clojures': [], 'endline': 17, 'complexity': 4, 'lineno': 13, 'is_method':
     False, 'name': 'aux', 'col_offset': 4, 'rank': 'A'},

    {'endline': 16, 'complexity': 10, 'lineno': 12, 'is_method': False, 'name':
     'name', 'col_offset': 0, 'rank': 'B'},
]


class TestDictConversion(unittest.TestCase):

    def test_raw_to_dict(self):
        self.assertEqual(tools.raw_to_dict(Module(103, 123, 98, 8, 19, 5)),
                         {'loc': 103, 'lloc': 123, 'sloc': 98, 'comments': 8,
                          'multi': 19, 'blank': 5})

    def test_cc_to_xml(self):
        self.assertEqual(tools.dict_to_xml({'filename': CC_TO_XML_CASE}),
                         '''<ccm>
                              <metric>
                                <complexity>6</complexity>
                                <unit>name</unit>
                                <classification>B</classification>
                                <file>filename</file>
                              </metric>
                              <metric>
                                <complexity>8</complexity>
                                <unit>Classname</unit>
                                <classification>B</classification>
                                <file>filename</file>
                              </metric>
                              <metric>
                                <complexity>7</complexity>
                                <unit>Classname.name</unit>
                                <classification>B</classification>
                                <file>filename</file>
                              </metric>
                              <metric>
                                <complexity>4</complexity>
                                <unit>aux</unit>
                                <classification>A</classification>
                                <file>filename</file>
                              </metric>
                              <metric>
                                <complexity>10</complexity>
                                <unit>name</unit>
                                <classification>B</classification>
                                <file>filename</file>
                              </metric>
                            </ccm>'''.replace('\n', '').replace(' ', ''))


CC_TO_TERMINAL_CASES = [
    Class(name='Classname', lineno=17, col_offset=0, endline=29,
          methods=[Function(name='meth', lineno=19, col_offset=4, endline=26,
                            is_method=True, classname='Classname', clojures=[],
                            complexity=3)], real_complexity=3),
    Function(name='meth', lineno=19, col_offset=4, endline=26, is_method=True,
             classname='Classname', clojures=[], complexity=7),
    Function(name='f1', lineno=12, col_offset=0, endline=16, is_method=False,
             classname=None, clojures=[], complexity=14),
    Function(name='f2', lineno=12, col_offset=0, endline=16, is_method=False,
             classname=None, clojures=[], complexity=22),
    Function(name='f3', lineno=12, col_offset=0, endline=16, is_method=False,
             classname=None, clojures=[], complexity=32),
    Function(name='f4', lineno=12, col_offset=0, endline=16, is_method=False,
             classname=None, clojures=[], complexity=41),
]


class TestCCToTerminal(unittest.TestCase):

    def test_cc_to_terminal(self):
        # do the patching
        tools.LETTERS_COLORS = dict((l, '<!{0}!>'.format(l)) for l in 'FMC')
        tools.RANKS_COLORS = dict((r, '<|{0}|>'.format(r)) for r in 'ABCDEF')
        tools.BRIGHT = '@'
        tools.RESET = '__R__'

        results = CC_TO_TERMINAL_CASES
        res = [
            '@<!C!>C __R__17:0 Classname - <|A|>A (4)__R__',
            '@<!M!>M __R__19:4 Classname.meth - <|B|>B (7)__R__',
            '@<!F!>F __R__12:0 f1 - <|C|>C (14)__R__',
            '@<!F!>F __R__12:0 f2 - <|D|>D (22)__R__',
            '@<!F!>F __R__12:0 f3 - <|E|>E (32)__R__',
            '@<!F!>F __R__12:0 f4 - <|F|>F (41)__R__',
        ]
        res_noshow = ['{0}__R__'.format(r[:r.index('(') - 1]) for r in res]

        self.assertEqual(tools.cc_to_terminal(results, False, 'A', 'F', False),
                         (res_noshow, 120, 6))
        self.assertEqual(tools.cc_to_terminal(results, True, 'A', 'F', False),
                         (res, 120, 6))
        self.assertEqual(tools.cc_to_terminal(results, True, 'A', 'D', False),
                         (res[:-2], 47, 4))
        self.assertEqual(tools.cc_to_terminal(results, False, 'A', 'D', False),
                         (res_noshow[:-2], 47, 4))
        self.assertEqual(tools.cc_to_terminal(results, True, 'C', 'F', False),
                         (res[2:], 109, 4))
        self.assertEqual(tools.cc_to_terminal(results, True, 'B', 'E', False),
                         (res[1:-1], 75, 4))
        self.assertEqual(tools.cc_to_terminal(results, True, 'B', 'F', True),
                         (res[1:], 120, 6))
